package com.iden.common.facetool;

import cn.hutool.core.collection.CollUtil;
import com.face.monitor.FaceMonitor;
import com.face.monitor.model.FaceModel;
import com.face.monitor.model.FaceRecogRetrieveResult;
import com.iden.common.cache.FaceMoniterCache;
import com.iden.common.constant.Constants;
import com.iden.common.facetool.pool.MyFaceMonitor;
import com.iden.common.facetool.pool.ResourcePoolableObjectFactory;
import com.iden.common.util.ByteUtil;
import com.iden.common.vo.FaceRetrieveResultVO;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.pool2.PooledObjectFactory;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class FaceIdenTool {
    private static final Logger logger = LogManager.getLogger(FaceIdenTool.class);

    public static void initPool() {
        // 创建池对象工厂
        PooledObjectFactory<MyFaceMonitor> factory = new ResourcePoolableObjectFactory();

        GenericObjectPoolConfig poolConfig = new GenericObjectPoolConfig();
        // 最大空闲数
        poolConfig.setMaxIdle(10);
        // 最小空闲数, 池中只有一个空闲对象的时候，池会在创建一个对象，并借出一个对象，从而保证池中最小空闲数为1
        poolConfig.setMinIdle(1);
        // 最大池对象总数
        poolConfig.setMaxTotal(100);
        // 逐出连接的最小空闲时间 默认1800000毫秒(30分钟)
        poolConfig.setMinEvictableIdleTimeMillis(1800000);
        // 逐出扫描的时间间隔(毫秒) 如果为负数,则不运行逐出线程, 默认-1
        poolConfig.setTimeBetweenEvictionRunsMillis(-1);
        // 在获取对象的时候检查有效性, 默认false
        poolConfig.setTestOnBorrow(true);
        // 在归还对象的时候检查有效性, 默认false
        poolConfig.setTestOnReturn(false);
        // 在空闲时检查有效性, 默认false
        poolConfig.setTestWhileIdle(false);
        // 最大等待时间， 默认的值为-1，表示无限等待。
        poolConfig.setMaxWaitMillis(5000);
        // 是否启用后进先出, 默认true
        poolConfig.setLifo(true);
        // 连接耗尽时是否阻塞, false报异常,ture阻塞直到超时, 默认true
        poolConfig.setBlockWhenExhausted(true);
        // 每次逐出检查时 逐出的最大数目 默认3
        poolConfig.setNumTestsPerEvictionRun(3);

        // 创建对象池
        final GenericObjectPool<MyFaceMonitor> pool = new GenericObjectPool<MyFaceMonitor>(factory, poolConfig);
        FaceMoniterCache.setGenericObjectPool(Constants.FACE_MONITOR_POOL_KEY, pool);
    }

    /**
     * 获取引擎
     * @return
     */
    public static MyFaceMonitor getMyFaceMonitor() {
        try {
            // 对象池
            final GenericObjectPool<MyFaceMonitor> pool = FaceMoniterCache.getGenericObjectPool(Constants.FACE_MONITOR_POOL_KEY);
            MyFaceMonitor myFaceMonitor = pool.borrowObject();
            logger.info("myFaceMonitor=="+myFaceMonitor);
            logger.info("FaceMonitor=="+myFaceMonitor.getFaceMonitor());
            return myFaceMonitor;
        } catch (Exception e){
            e.printStackTrace();
        }
        return null;
    }


    /**
     * 批量获取人脸特征码结构体
     * @param imgFiles
     * @return
     */
    public static FaceModel[] extractFeature(File[] imgFiles){
        //获取引擎
        MyFaceMonitor myFaceMonitor = getMyFaceMonitor();
        FaceMonitor faceMonitor = myFaceMonitor.getFaceMonitor();

        List<byte[]> faceImageList = new ArrayList<>();
        if(imgFiles != null && imgFiles.length > 0){
            for(File imgFile : imgFiles) {
                faceImageList.add(readFileBytes(imgFile.getAbsolutePath()));
            }
        }
        FaceModel[] faceModels = null;
        if (CollUtil.isNotEmpty(faceImageList)){
            faceMonitor.extractFeature(faceImageList);
        }

        returnObjectToPool(myFaceMonitor);
        return faceModels;
    }


    /**
     * 得到单个特征码
     * @param imgFile
     * @return
     */
    public static String getFeatPtr(File imgFile) {
        if(imgFile == null || !imgFile.isFile()){
            return null;
        }
        File[] imgFiles = new File[]{imgFile};
        FaceModel[] faceModels = extractFeature(imgFiles);

        if(faceModels != null && faceModels.length > 0) {
            FaceModel faceModel = faceModels[0];
            logger.info("name; " + faceModel.getName());
            logger.info("personId: " + faceModel.getPersonId());
            return ByteUtil.byte2Hex(faceModel.getFeatValue());
        }
        return null;
    }


    /**
     * faceModel在faceMonitor的人脸特征库中检索命中的底库的index
     * @param faceMonitor
     * @param faceModel
     * @return
     */
    public static FaceRetrieveResultVO getHitResult(FaceMonitor faceMonitor, FaceModel faceModel) {
        if (faceMonitor == null){
            return null;
        }
        FaceRecogRetrieveResult faceRecogRetrieveResult = faceMonitor.faceRecogRetrieveFaceDataset(faceModel);
        if(faceRecogRetrieveResult != null) {
            int index = faceRecogRetrieveResult.getOnePredictIds()[0];
            float score = faceRecogRetrieveResult.getOnePredictScores()[0];
            if (score >= 65) {
                FaceRetrieveResultVO faceRetrieveResultVO = new FaceRetrieveResultVO();
                faceRetrieveResultVO.setIndex(index);
                faceRetrieveResultVO.setScore(score);
                return faceRetrieveResultVO;
            }
        }
        return null;
    }
    public static FaceRetrieveResultVO getHitResult(FaceMonitor faceMonitor,String featPtrVisitor){
        FaceModel faceModelPtrVisitor = new FaceModel();
        faceModelPtrVisitor.setName(String.valueOf(0));
        faceModelPtrVisitor.setPersonId(0);
        logger.info("FaceIdenTool getHitResult2 3...");
        faceModelPtrVisitor.setFeatValue(ByteUtil.hex2Byte(featPtrVisitor));
        logger.info("FaceIdenTool getHitResult2 4..");
        FaceRetrieveResultVO vo = getHitResult(faceMonitor,faceModelPtrVisitor);
        logger.info("FaceIdenTool getHitResult2 5..");
        return vo;
    }

    /**
     * 归还人脸引擎到池子里
     * @param myFaceMonitor
     */
    public static void returnObjectToPool(MyFaceMonitor myFaceMonitor) {
        // 对象池
        final GenericObjectPool<MyFaceMonitor> pool = FaceMoniterCache.getGenericObjectPool(Constants.FACE_MONITOR_POOL_KEY);

        pool.returnObject(myFaceMonitor);// 申请的资源用完了记得归还，不然其他人要申请时可能就没有资源用了

    }

    /**
     * 释放人脸引擎
     * @param faceMonitor
     */
    public static void releaseEngine(FaceMonitor faceMonitor) {
        if(faceMonitor != null){
            faceMonitor.releaseEngine();
            faceMonitor = null;
        }
    }
    /**
     * read file
     *
     * @param filePath
     * @return if file not exist, return null, else return content of file
     * @throws RuntimeException if an error occurs while operator BufferedReader
     */
    private static byte[] readFileBytes(String filePath) {
        File file = new File(filePath);
        if (file == null || !file.isFile()) {
            return null;
        }

        InputStream inputStream = null;
        ByteArrayOutputStream baos = null;
        try {
            inputStream = new FileInputStream(file);
            baos = new ByteArrayOutputStream();

            byte[] buffer = new byte[1024];
            int byteCount = 0;
            while ((byteCount = inputStream.read(buffer)) != -1) {// 循环从输入流读取
                // buffer字节
                baos.write(buffer, 0, byteCount);// 将读取的输入流写入到输出流
            }
            baos.flush();// 刷新缓冲区
            return baos.toByteArray();
        } catch (IOException e) {
            return null;
            //throw new RuntimeException("IOException occurred. ", e);
        } finally {
            close(inputStream);
            close(baos);
        }
    }

    private static void close(Closeable closeable) {
        if (closeable != null) {
            try {
                closeable.close();
            } catch (IOException e) {
                throw new RuntimeException("IOException occurred. ", e);
            }
        }
    }


    public static boolean isBad(String featPtr){
        if (StringUtils.isEmpty(featPtr)) {
            return true;
        }
        char[] chars = featPtr.toCharArray();

        for (char c : chars){
            if(!String.valueOf(c).equals("0")){
                return false;
            }
        }
        return true;
    }
}
